import lldb
from pydoc import locate
from lldb_rb.constants import *
from lldb_rb.utils import *

class RbBaseCommand(LLDBInterface):
    @classmethod
    def register_lldb_command(cls, debugger, module_name):
        # Add any commands contained in this module to LLDB
        command = f"command script add -c {module_name}.{cls.__name__} {cls.program}"
        debugger.HandleCommand(command)

    @classmethod
    def lldb_init(cls, debugger):
        target = debugger.GetSelectedTarget()
        global SIZEOF_VALUE
        SIZEOF_VALUE = target.FindFirstType("VALUE").GetByteSize()

        value_types = []
        g = globals()

        imemo_types = target.FindFirstType("enum imemo_type")

        #for member in imemo_types.GetEnumMembers():
        #    g[member.GetName()] = member.GetValueAsUnsigned()

        for enum in target.FindFirstGlobalVariable("ruby_dummy_gdb_enums"):
            enum = enum.GetType()
            members = enum.GetEnumMembers()
            for i in range(0, members.GetSize()):
                member = members.GetTypeEnumMemberAtIndex(i)
                name = member.GetName()
                value = member.GetValueAsUnsigned()
                g[name] = value

                if name.startswith("RUBY_T_"):
                    value_types.append(name)
        g["value_types"] = value_types
        return g

    def __init__(self, debugger, _internal_dict):
        self.ruby_globals = RbBaseCommand.lldb_init(debugger)
        self.internal_dict = _internal_dict

    def __call__(self, debugger, command, exe_ctx, result):
        self.ruby_globals = RbBaseCommand.lldb_init(debugger)
        self.build_environment(debugger)
        self.call(debugger, command, exe_ctx, result)

    def call(self, debugger, command, exe_ctx, result):
        raise NotImplementedError("subclasses must implement call")

    def get_short_help(self):
        return self.__class__.help_string

    def get_long_help(self):
        return self.__class__.help_string



import lldb
import re

from lldb_rb.constants import *
from lldb_rb.rb_base_command import RbBaseCommand

class PrintFlagsCommand(RbBaseCommand):
    program = "print_flags"

    help_string = "Print out the individial flags of an RVALUE object in human readable format"

    # call is where our command logic will be implemented
    def call(self, debugger, command, exe_ctx, result):
        rclass_t = self.target.FindFirstType("::RBasic")
        rcass_ptr = self.target.EvaluateExpression(command).Cast(rclass_t.GetPointerType())
        obj_flags = rcass_ptr.GetValueForExpressionPath("->flags").GetValueAsUnsigned()

        flags = [
            "RUBY_FL_WB_PROTECTED", "RUBY_FL_PROMOTED", "RUBY_FL_FINALIZE",
            "RUBY_FL_SHAREABLE", "RUBY_FL_FREEZE",
            "RUBY_FL_USER0", "RUBY_FL_USER1", "RUBY_FL_USER2", "RUBY_FL_USER3", "RUBY_FL_USER4",
            "RUBY_FL_USER5", "RUBY_FL_USER6", "RUBY_FL_USER7", "RUBY_FL_USER8", "RUBY_FL_USER9",
            "RUBY_FL_USER10", "RUBY_FL_USER11", "RUBY_FL_USER12", "RUBY_FL_USER13", "RUBY_FL_USER14",
            "RUBY_FL_USER15", "RUBY_FL_USER16", "RUBY_FL_USER17", "RUBY_FL_USER18"
        ]

        types_index = {v: k for k, v in self.ruby_globals.items() if re.match(r'RUBY_T_', k)}
        print("TYPE: {}".format(types_index[obj_flags & self.ruby_globals["RUBY_T_MASK"]]))
        for flag in flags:
            output = "{} : {}".format(flag, "1" if (obj_flags & self.ruby_globals[flag]) else "0")
            print(output, file=result)



class LLDBInterface:
    def build_environment(self, debugger):
        self.debugger = debugger
        self.target = debugger.GetSelectedTarget()
        self.process = self.target.GetProcess()
        self.thread = self.process.GetSelectedThread()
        self.frame = self.thread.GetSelectedFrame()

    def _append_command_output(self, command):
        output1 = self.result.GetOutput()
        self.debugger.GetCommandInterpreter().HandleCommand(command, self.result)
        output2 = self.result.GetOutput()
        self.result.Clear()
        self.result.write(output1)
        self.result.write(output2)

    def _append_expression(self, expression):
        self._append_command_output("expression " + expression)
